import atexit
import json
import logging
import os
import queue
import threading
import typing
import uuid
from datetime import datetime

from mlfoundry.exceptions import MlFoundryException
from mlfoundry.monitoring.entities import (
    Actual,
    ActualPacket,
    BasePacket,
    Prediction,
    PredictionPacket,
)
from mlfoundry.monitoring.store.constants import (
    FLUSH_BATCH_SIZE_ENV_VAR,
    FLUSH_INTERVAL_ENV_VAR,
    MAX_QUEUE_SIZE_ENV_VAR,
)
from mlfoundry.monitoring.store.repositories import get_monitoring_store
from mlfoundry.monitoring.store.worker import MonitoringStoreWorker

logger = logging.getLogger(__name__)


def _check_shutdown(func):
    def wrapper(self, *args, **kwargs):
        if self._is_shutdown:
            raise MlFoundryException("cannot execute after client is shutdown")
        return func(self, *args, **kwargs)

    return wrapper


class MonitoringClient:
    def __init__(self, monitoring_uri: str, model_version_id: str):
        self._is_shutdown = False

        self.monitoring_store = get_monitoring_store(host_uri=monitoring_uri)
        self.model_version_id = model_version_id
        self.monitoring_store.enable_monitoring_if_not_enabled(
            model_version_id=model_version_id
        )
        max_queue_size = int(os.getenv(MAX_QUEUE_SIZE_ENV_VAR, "10000"))
        flush_interval = float(os.getenv(FLUSH_INTERVAL_ENV_VAR, "5"))
        flush_batch_size = int(os.getenv(FLUSH_BATCH_SIZE_ENV_VAR, "300"))

        if max_queue_size <= 0:
            raise MlFoundryException(
                f"{MAX_QUEUE_SIZE_ENV_VAR} should be a positive number"
            )

        self.task_queue: "queue.Queue[BasePacket]" = queue.Queue(max_queue_size)
        self.worker_terminate_event = threading.Event()

        if flush_interval <= 0:
            raise MlFoundryException(
                f"{FLUSH_INTERVAL_ENV_VAR} should be a positive number"
            )
        if flush_batch_size <= 0 or flush_batch_size > max_queue_size:
            raise MlFoundryException(
                f"{FLUSH_BATCH_SIZE_ENV_VAR} should be positive"
                f" and lower than {MAX_QUEUE_SIZE_ENV_VAR}"
            )

        self.worker: MonitoringStoreWorker = MonitoringStoreWorker(
            task_queue=self.task_queue,
            terminate_event=self.worker_terminate_event,
            monitoring_store=self.monitoring_store,
            flush_interval=flush_interval,
            flush_every_num_message=flush_batch_size,
        )
        atexit.register(self._shutdown)
        self.worker.start()

    @staticmethod
    def generate_hash_from_data(
        features: typing.Dict, timestamp: typing.Optional[datetime] = None
    ):
        data_str = json.dumps(features, sort_keys=True)
        if timestamp:
            data_str += timestamp.isoformat()
        return uuid.uuid5(uuid.NAMESPACE_X500, data_str).hex

    @staticmethod
    def generate_random_id():
        return uuid.uuid4().hex

    def _generate_prediction_packet_from_data(
        self, prediction: Prediction
    ) -> PredictionPacket:

        return PredictionPacket(
            model_version_id=self.model_version_id,
            prediction=prediction,
        )

    def _generate_actual_packet_from_data(self, actual: Actual):
        return ActualPacket(
            model_version_id=self.model_version_id,
            actual=actual,
        )

    def _put_in_queue(self, inference_packet: BasePacket):
        try:
            self.task_queue.put_nowait(inference_packet)
        except queue.Full as ex:
            raise MlFoundryException(
                "task queue is full\n"
                f"current task queue length is {self.task_queue.maxsize}\n"
                "consider increasing the task queue length using "
                f"{MAX_QUEUE_SIZE_ENV_VAR} environment variable"
            ) from ex

    @_check_shutdown
    def log_prediction(self, prediction: Prediction):
        prediction_packet = self._generate_prediction_packet_from_data(prediction)
        self._put_in_queue(prediction_packet)

    @_check_shutdown
    def log_actual(self, actual: Actual):
        actual_packet = self._generate_actual_packet_from_data(actual)
        self._put_in_queue(actual_packet)

    @_check_shutdown
    def log_predictions(self, predictions: typing.List[Prediction]):
        for prediction in predictions:
            self.log_prediction(prediction)

    @_check_shutdown
    def log_actuals(self, actuals: typing.List[Actual]):
        for actual in actuals:
            self.log_actual(actual)

    def _flush(self):
        logger.debug(
            f"flushing task queue, {self.task_queue.qsize()} items in the queue"
        )
        self.task_queue.join()
        logger.debug("task queue flushed")

    def _shutdown(self):
        if self._is_shutdown:
            return
        logger.debug("shutting down worker and client")
        self._is_shutdown = True
        # NOTE: We initialize the monitoring store at first in the constructor
        # The task_queue, worker is defined later.
        # There is a chance that monitoring  store initialization will throw error,
        # in that case, shutdown will be called (__del__) but self.task_queue would not have
        # been initialized yet.
        if hasattr(self, "task_queue"):
            self._flush()
        if hasattr(self, "worker_terminate_event"):
            logger.debug("setting worker termination event")
            self.worker_terminate_event.set()
        if hasattr(self, "worker"):
            logger.debug("waiting for worker to terminate")
            self.worker.join()

    def __del__(self):
        self._shutdown()



import logging
import typing
from datetime import datetime

from pydantic import BaseModel, Field, constr

logger = logging.getLogger(__name__)

# TODO @nikp1172 Add support for np types https://numpy.org/doc/stable/user/basics.types.html
EntityValue = typing.Union[str, int, float]


class PredictionData(BaseModel):
    value: EntityValue
    probabilities: typing.Dict[str, float] = Field(default_factory=dict)
    shap_values: typing.Dict[str, float] = Field(default_factory=dict)

    class Config:
        smart_union = True


class BasePacket(BaseModel):
    class Config:
        smart_union = True


class Prediction(BaseModel):

    data_id: constr(min_length=1, max_length=64)
    features: typing.Dict[str, EntityValue]
    prediction_data: PredictionData

    raw_data: typing.Optional[typing.Dict] = Field(default_factory=dict)
    occurred_at: datetime = Field(
        default_factory=lambda: datetime.utcnow().replace(tzinfo=None)
    )

    class Config:
        smart_union = True


class PredictionPacket(BasePacket):
    model_version_id: constr(min_length=1, max_length=64)
    prediction: Prediction


class Actual(BaseModel):
    data_id: constr(min_length=1, max_length=64)
    value: EntityValue

    class Config:
        smart_union = True


class ActualPacket(BasePacket):
    model_version_id: constr(min_length=1, max_length=64)
    actual: Actual



import typing
from collections import defaultdict

from mlflow.tracking._tracking_service.utils import _get_default_host_creds
from mlflow.utils.rest_utils import MlflowHostCreds, http_request_safe

from mlfoundry.monitoring.entities import ActualPacket, BasePacket, PredictionPacket
from mlfoundry.monitoring.store.repositories.dto import (
    ActualData,
    BatchInsertRequest,
    BatchUpdateActualRequest,
    ClassPrediction,
    Data,
    MlModelPrediction,
)
from mlfoundry.tracking.entities import AuthServerInfo


# TODO @nikp1172 add feature of autmatically serializing datetime object in \
#  mlflow.utils.rest_utils.http_request
class RestMonitoringStore:
    def __init__(self, host_uri: str):
        self.host_creds = _get_default_host_creds(store_uri=host_uri)

    def _insert_prediction(self, prediction_packets: typing.List[PredictionPacket]):
        if not prediction_packets:
            return
        items = [
            Data(
                data_id=packet.prediction.data_id,
                features=packet.prediction.features,
                prediction=MlModelPrediction(
                    value=packet.prediction.prediction_data.value,
                    probabilities=[
                        ClassPrediction(label=feature_name, score=score)
                        for feature_name, score in packet.prediction.prediction_data.probabilities.items()
                    ],
                    shap_values=packet.prediction.prediction_data.shap_values,
                    occurred_at=packet.prediction.occurred_at,
                ),
                raw_data=packet.prediction.raw_data,
            )
            for packet in prediction_packets
        ]
        batch_insert_request = BatchInsertRequest(
            model_version_id=prediction_packets[0].model_version_id,
            items=items,
        )

        response = http_request_safe(
            host_creds=self.host_creds,
            endpoint="/v1/data/batch-insert",
            method="post",
            json=batch_insert_request.to_json_dict(),
        )

    def _update_actuals(self, actual_packets: typing.List[ActualPacket]):
        if not actual_packets:
            return

        batch_update_request = BatchUpdateActualRequest(
            model_version_id=actual_packets[0].model_version_id,
            items=[
                ActualData(data_id=packet.actual.data_id, actual=packet.actual.value)
                for packet in actual_packets
            ],
        )
        response = http_request_safe(
            host_creds=self.host_creds,
            endpoint="/v1/data/batch-update-actual",
            method="post",
            json=batch_update_request.dict(),
        )

    def batch_log_inference(self, inference_packets: typing.List[BasePacket]):
        grouped_packets = defaultdict(list)
        for packet in inference_packets:
            grouped_packets[type(packet)].append(packet)

        # log predictions before actuals to avoid actuals getting logged before predictions
        if PredictionPacket in grouped_packets.keys():
            self._insert_prediction(grouped_packets[PredictionPacket])
        if ActualPacket in grouped_packets.keys():
            self._update_actuals(grouped_packets[ActualPacket])

    def enable_monitoring_if_not_enabled(self, model_version_id: str):
        response = http_request_safe(
            host_creds=self.host_creds,
            endpoint=f"/v1/model/enable",
            method="post",
            params={"model_version_id": model_version_id},
        )

    @staticmethod
    def get_auth_server_info(monitoring_uri: str) -> AuthServerInfo:
        response = http_request_safe(
            host_creds=MlflowHostCreds(host=monitoring_uri),
            endpoint="/v1/model/auth-server-info",
            method="get",
        )
        response = response.json()
        return AuthServerInfo.parse_obj(response)
